/*
 * Teensy RC airplane hub. 
 * 
 * Made for the Teensy 3.2. The microcontroller reads an S-BUS input with 16 channals. 
 * Up to 10 channals can be used to control servo PWM outputs. All servo outputs can
 * be a mix of diffrent input channals, and can also mixed with one or more gyro
 * axis. An MPU-6050 gyro accelerometer is used. The microcontroller also controls
 * LED lights trough MOTFET outputs. Navigationlights, two landing light channels, 
 * one beacon light and one stobe light can be configured and controled with the
 * S-BUS input signals. Servo outputs, mixing and light controls can be configured
 * using a simple interface on the USB serial port. 
 * 
 * by Axel Brinkeby
 * 2019-04-14
 * 
 */

#include "SBUS.h"   // https://github.com/bolderflight/SBUS
#include <Servo.h>  // https://www.pjrc.com/teensy/td_libs_Servo.html
#include "Pwmoutput.h"
#include "mpu6050.h"

// pins: 
const int NAV_LIGHTS = 23;
const int STROBE_LIGHTS = 22;
const int LANDING_LIGHT_1 = 21;
const int LANDING_LIGHT_2 = 20;
const int BEACON_LIGHT = 6;
const int PWM_1 = 2;
const int PWM_2 = 3;
const int PWM_3 = 4;
const int PWM_4 = 5;
const int PWM_5 = 7;
const int PWM_6 = 8;
const int PWM_7 = 11;
const int PWM_8 = 15;
const int PWM_9 = 16;
const int PWM_10 = 17;
const int ARM_SW = 12;
const int STATUS_LED = 13;

// SBUS input channels: 
const int CH_TROTTLE = 0;         // channel 1
const int CH_AILRON_1 = 1;        // channel 2
const int CH_ELEVATOR_1 = 2;      // channel 3
const int CH_RUDDER = 3;          // channel 4
const int CH_AILRON_2 = 4;        // channel 5
const int CH_FLAP_1 = 5;          // channel 6
const int CH_FLAP_2 = 6;          // channel 7
const int CH_ELEVATOR_2 = 7;      // channel 8
const int CH_NAV_LIGHTS = 8;      // channel 9
const int CH_LANDING_LIGHTS = 9;  // channel 10
const int CH_STOBE_LIGHTS = 10;   // channel 11
const int CH_BEACON_LIGHTS = 11;  // channel 12
const int CH_ARM = 12;            // channel 13
const int CH_GYRO_GAIN = 13;      // channel 14
const int CH_NOT_USED_3 = 14;     // channel 15
const int CH_NOT_USED_4 = 15;     // channel 16

const int GYRO_ROLL_INDEX = 16;   // used to set gyro mix factors in the PWM-output class. 
const int GYRO_PITCH_INDEX = 17;     
const int GYRO_YAW_INDEX = 18;  

// PWM out functions: 
const int LEFT_AILRON = 0;    // PWM output 1
const int RIGHT_AILRON = 1;   // PWM output 2
const int LEFT_FLAP = 2;      // PWM output 3
const int RIGHT_FLAP = 3;     // PWM output 4
const int MOTOR = 4;          // PWM output 5
const int LEFT_ELEVATOR = 5;  // PWM output 6
const int RUDDER = 6;         // PWM output 7
const int RIGHT_ELEVATOR = 7; // PWM output 8
const int NOT_UESD_1 = 8;     // PWM output 9
const int NOT_UESD_2 = 9;     // PWM output 10

Mpu6050 imu(0x68);  // default adress (on the common breakout-board) is 0x69
Pwmoutput pwmOutputs[10];
SBUS sbus(Serial1);

uint16_t channels[16];  // "channels" is raw SBUS channel data, ranging from 0 to 2000 per channel, 16 channels
bool failSafe;
bool lostFrame;
float gyroValues[3];

int beaconPeriod = 2000;
int beaconOffset = 0; 
int strobePeriod = beaconPeriod;
int strobeOffset = strobePeriod/2; 
int strobeDuration = 60; 
int wigWagPeriod = 1400;

bool strobeOn = false; 
bool navLightsOn = false; 
bool beaconOn = false;
bool landingLightsOn = false; 
bool landingLightsWigWag = false;

void setup() {
  Serial.begin(115200); // USB is always 12 Mbit/sec
  Serial.println("Teensy RC airplane hub");
  Serial.println("by Axel Brinkeby");
  Serial.println("V 1.0 \n");

  pinMode(NAV_LIGHTS, OUTPUT);
  digitalWrite(NAV_LIGHTS, false);
  pinMode(STROBE_LIGHTS, OUTPUT);
  digitalWrite(STROBE_LIGHTS, false);
  pinMode(LANDING_LIGHT_1, OUTPUT);
  digitalWrite(LANDING_LIGHT_1, false);
  pinMode(LANDING_LIGHT_2, OUTPUT);
  digitalWrite(LANDING_LIGHT_2, false);
  pinMode(BEACON_LIGHT, OUTPUT);
  digitalWrite(BEACON_LIGHT, false);

  pinMode(STATUS_LED, OUTPUT);
  digitalWrite(STATUS_LED, false);  

  pinMode(ARM_SW, INPUT_PULLUP);

  pwmOutputs[LEFT_AILRON].attachOnPin(PWM_1);
  pwmOutputs[LEFT_AILRON].setInput(CH_AILRON_1, 1.0); 
  pwmOutputs[LEFT_AILRON].setInput(GYRO_ROLL_INDEX, 1.0); 
  
  pwmOutputs[RIGHT_AILRON].attachOnPin(PWM_2);
  pwmOutputs[RIGHT_AILRON].setInput(CH_AILRON_2, 1.0);
  pwmOutputs[RIGHT_AILRON].setInput(GYRO_ROLL_INDEX, 1.0);
  
  pwmOutputs[LEFT_FLAP].attachOnPin(PWM_3);
  pwmOutputs[LEFT_FLAP].setInput(CH_FLAP_1, 1.0);
  pwmOutputs[LEFT_FLAP].setInput(CH_AILRON_1, 0.5);
  
  pwmOutputs[RIGHT_FLAP].attachOnPin(PWM_4);
  pwmOutputs[RIGHT_FLAP].setInput(CH_FLAP_2, -1.0);
  pwmOutputs[RIGHT_FLAP].setInput(CH_AILRON_2, 0.5);
  
  pwmOutputs[MOTOR].attachOnPin(PWM_5);
  pwmOutputs[MOTOR].setInput(CH_TROTTLE, 1.0);
  
  pwmOutputs[LEFT_ELEVATOR].attachOnPin(PWM_6);
  pwmOutputs[LEFT_ELEVATOR].setInput(CH_ELEVATOR_1, -1.0);
  pwmOutputs[LEFT_ELEVATOR].setInput(CH_AILRON_1, -0.3);
  //pwmOutputs[LEFT_ELEVATOR].setInput(CH_FLAP_1, -0.15);
  pwmOutputs[LEFT_ELEVATOR].setInput(GYRO_PITCH_INDEX, -1.0);
  
  pwmOutputs[RUDDER].attachOnPin(PWM_7);
  pwmOutputs[RUDDER].setInput(CH_RUDDER, -1.0);
  pwmOutputs[RUDDER].setInput(GYRO_YAW_INDEX, -1.0);
  
  pwmOutputs[RIGHT_ELEVATOR].attachOnPin(PWM_8);
  pwmOutputs[RIGHT_ELEVATOR].setInput(CH_ELEVATOR_2, 1.0);
  pwmOutputs[RIGHT_ELEVATOR].setInput(CH_AILRON_2, -0.3);
  //pwmOutputs[RIGHT_ELEVATOR].setInput(CH_FLAP_2, 0.15);
  pwmOutputs[RIGHT_ELEVATOR].setInput(GYRO_PITCH_INDEX, 1.0);
  
  // pwmOutputs[NOT_UESD_1].attachOnPin(PWM_9);   // not used
  // pwmOutputs[NOT_UESD_2].attachOnPin(PWM_10);  // not used  

  digitalWrite(BEACON_LIGHT, true);
  Wire.begin();                         
  Wire.setClock(100000);
  imu.init(); 
  imu.calibradeGyro();
  digitalWrite(BEACON_LIGHT, false);

  sbus.begin();
}

void loop() {
  
  // look for a good SBUS packet from the receiver
  if(sbus.read(&channels[0], &failSafe, &lostFrame)){

    imu.updateIMUdata();
    gyroValues[0] = imu.getGyroX();
    gyroValues[1] = imu.getGyroY();
    gyroValues[2] = imu.getGyroZ();
    //Serial.println(gyroValues[0]);

    // Adjust the gyro gain for all gyro axis with the gyro gain channel. 
    float normalizedGyroGain = constrain((channels[CH_GYRO_GAIN] - 200.0f) / 1600.0f, 0.0f, 1.0f); 
    for (int i = 0; i < 3; i++) {
      gyroValues[i] *= normalizedGyroGain; 
    }    

    // motor is locked in centre position if the arm switch is not activated
    //if (digitalRead(ARM_SW)) {  // Hardware ARM switch
    if (channels[CH_ARM] < 750) {  // ARM channel from the transmitter
      channels[CH_TROTTLE] = 1000;   
    }
    
    digitalWrite(STATUS_LED, !digitalRead(ARM_SW) & millis() % 200 < 100);
    
    for (int i = 0; i < 8; i++) {
      pwmOutputs[i].setInputValues(channels, gyroValues);
      pwmOutputs[i].updateOutput();
    }
    
    navLightsOn = (channels[CH_NAV_LIGHTS] > 750); 
    landingLightsOn = (channels[CH_LANDING_LIGHTS] > 750); 
    landingLightsWigWag = (channels[CH_LANDING_LIGHTS] > 1250);
    strobeOn = (channels[CH_STOBE_LIGHTS] > 750); 
    beaconOn = (channels[CH_BEACON_LIGHTS] > 750);     
  }

  updateLights();  
}

void updateLights() {
  if (navLightsOn) {    // red and green lights on the wing tips
    digitalWrite(NAV_LIGHTS, true);
  } else {
    digitalWrite(NAV_LIGHTS, false);
  }

  if (strobeOn) {       // blight white lights on the wing tips
    bool flash1 = (millis() + strobeOffset) % strobePeriod < strobeDuration;
    bool flash2 = (millis() + strobeOffset + 120) % strobePeriod < strobeDuration;
    digitalWrite(STROBE_LIGHTS, flash1 || flash2);
  } else {
    digitalWrite(STROBE_LIGHTS, false);
  }

  byte leftLandingLightValue = 0, rightLandingLightValue = 0;   
  if (landingLightsOn) {      // one bright white landing light on each side on the wing
    if (landingLightsWigWag) {  // alternate left and right landing light, used for high visibility
      if (millis() % wigWagPeriod < wigWagPeriod/2) {
        leftLandingLightValue = 255; 
        rightLandingLightValue = 0; 
      } else {
        leftLandingLightValue = 0; 
        rightLandingLightValue = 255; 
      }      
    } else {
      leftLandingLightValue = rightLandingLightValue = 150; 
    }
  } 
  analogWrite(LANDING_LIGHT_1, leftLandingLightValue);
  analogWrite(LANDING_LIGHT_2, rightLandingLightValue);
    
  byte beaconLEDvalue = 0; 
  if (failSafe) {     // use the beacon as a warning that failsafe is triggered
    if(millis() % 200 < 50) {
      beaconLEDvalue = 255;
    } else {
      beaconLEDvalue = 0; 
    }
  } else if (beaconOn) {  // normal beacon flashing
    int temp = (millis() + beaconOffset + beaconPeriod/2) % beaconPeriod; 
    temp = max(min(abs(temp - (beaconPeriod/2)), 255), 0);
    beaconLEDvalue = 255 - temp;
  } 
  analogWrite(BEACON_LIGHT, beaconLEDvalue);
}
